// vim:tabstop=4:shiftwidth=4:expandtab:cinoptions=(s,U1,m1
// Copyright (C) 2005 Dmitry Poplavsky <dima@thekompany.com>

#ifndef CALLSLOG_H
#define CALLSLOG_H

#include <qobject.h>
#include <qdatetime.h>
#include <qvaluelist.h>

#include "callslogdialogbase.h"

class CallItem {
public:
    typedef enum { Incoming, Outgoing } Direction;

    CallItem();
    CallItem( QString data ); // load from serialized string
    CallItem( QString jid, QDateTime callTime, Direction direction ); // for rejected calls
    CallItem( QString jid, QDateTime callTime, Direction direction,
              int duration ); // for accepted calls

    QString toString();

    QString jid;
    QDateTime callTime;
    Direction direction;
    bool isAccepted;
    int duration; // in seconds

    bool isValid; 
};

/*!
 \class CallsLog callslog.h
 \brief stub
 */

class CallsLog : public QObject {
Q_OBJECT
public:
	CallsLog( QString fileName, QObject * parent = 0L, const char * name = 0L );
	virtual ~CallsLog();

    static CallsLog* instance();

    QValueList<CallItem> items();

public slots:
    void addItem( CallItem& );

private:
	class Private;
	Private *d;
}; 


class CallsLogDialog : public CallsLogDialogBase {
Q_OBJECT
public:
    CallsLogDialog( QWidget* parent = 0, const char* name = 0, bool modal = FALSE, WFlags fl = 0 );
    ~CallsLogDialog();

public slots:
    virtual void showOnlyAccepted(bool);
    virtual void showCallsWith( const QString& );

protected:
    void updateItems();
    
};

#endif // CALLSLOG_H


